// Database.cpp
//
// 20000818 / dhartmei - initial revision
// 20000905 / dhartmei - added db based database
// 20001001 / dhartmei - db disabled for Linux compatibility
//

#include "Database.hpp"
#include <fstream>

extern "C++" void syslog(std::string msg);

// ----------------------------------------------------------------------------

inline void toLower(std::string &s)
  {
    for (unsigned i = 0; i < s.length(); ++i)
      if ((s[i] >= 'A') && (s[i] <= 'Z')) s[i] = s[i]-'A'+'a';
    return;
  }

// ----------------------------------------------------------------------------

MemoryDatabase::MemoryDatabase(std::string filename) : Database(), filename(filename)
  {
  }

// ----------------------------------------------------------------------------

MemoryDatabase::~MemoryDatabase()
  {
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::clear()
  {
    syslog(std::string("--- clearing database ")+filename);
    entries.erase(entries.begin(), entries.end());
    save();
    return;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::load()
  {
    syslog(std::string("--- loading database ")+filename);
    importFile(filename);
    return;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::save() const
  {
    syslog(std::string("--- saving database ")+filename);
    exportFile(filename);
    return;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::importFile(std::string filename)
  {
    ifstream file;
    file.open(filename.c_str(), ios::in);
    if (file.good())
      {
        char line[8096];
        while (file.getline(line, 8095))
          {
            char *l = line; while (*l && (*l != '=')) ++l;
            if ((l == line) || !*l) syslog(std::string("MemoryDatabase::load(")+filename+") ignoring line without '='");
            else
              {
                *(l-1) = 0;
                l++; while (*l == ' ') l++;
                if (strlen(l) > 256) syslog(std::string("MemoryDatabase::load(")+filename+") ignoring overlong line");
                else insert(line, l);
              }
          }
        file.close();
      }
    return;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::exportFile(std::string filename) const
  {
    ofstream file;
    file.open(filename.c_str(), ios::out);
    if (file.good())
      {
        for (Map::const_iterator i = entries.begin(); i != entries.end(); ++i)
          {
            std::string key = (*i).first, value = (*i).second;
            toLower(key);
            file << key << " = " << value << endl;
          }
        file.flush();
        file.close();
      }
    return;
  }

// ----------------------------------------------------------------------------

std::string MemoryDatabase::find(std::string key) const
  {
    Map::const_iterator i = entries.find(key);
    if (i == entries.end()) return ""; else return (*i).second;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::insert(std::string key, std::string value)
  {
    entries.erase(key);
    entries.insert(Map::value_type(key, value));
    return;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::remove(std::string key)
  {
    entries.erase(key);
    return;
  }

// ----------------------------------------------------------------------------

void MemoryDatabase::apropos(std::vector<std::string> words, std::vector<std::string> &keys) const
  {
    for (Map::const_iterator i = entries.begin(); i != entries.end(); ++i)
      {
        std::string txt = (*i).first+" "+(*i).second; toLower(txt);
        bool ok = true;
        for (std::vector<std::string>::const_iterator j = words.begin(); ok && (j != words.end()); ++j)
          {
            if ((*j)[0] == '!') ok = (txt.find((*j).c_str()+1) == std::string::npos);
            else                ok = (txt.find((*j).c_str()  ) != std::string::npos);
          }
        if (ok) keys.push_back((*i).first);
      }
    return;
  }

// ----------------------------------------------------------------------------

#if (__WIN32__ || __Linux__)
  #warning FileDatabase skipped on WIN32 and Linux
#else

FileDatabase::FileDatabase(std::string filename) : Database(), filename(filename)
  {
    btreeinfo.flags      = 0;
    btreeinfo.cachesize  = 0;
    btreeinfo.maxkeypage = 0;
    btreeinfo.minkeypage = 0;
    btreeinfo.psize      = 0;
    btreeinfo.compare    = 0;
    btreeinfo.prefix     = 0;
    btreeinfo.lorder     = 0;
    open();
  }

// ----------------------------------------------------------------------------

FileDatabase::~FileDatabase()
  {
    close();
  }

// ----------------------------------------------------------------------------

void FileDatabase::open()
  {
    db = dbopen(filename.c_str(), O_CREAT|O_RDWR, 0600, DB_BTREE, &btreeinfo);
    if (!db) syslog("FileDatabase::open() dbopen() failed");
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::close()
  {
    if (db) if (db->close(db)) syslog("FileDatabase::close() close() failed");
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::clear()
  {
    close();
    if (unlink(filename.c_str())) syslog(std::string("FileDatabase::clear() couldn't delete file ")+filename);
    open();
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::save() const
  {
    if (!db) return;
    int r = db->sync(db, 0);
    if (r) syslog("FileDatabase::save() sync() failed");
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::importFile(std::string filename)
  {
    if (!db) return;
    ifstream file;
    file.open(filename.c_str(), ios::in);
    if (!file.good()) syslog(std::string("FileDatabase::importFile() couldn't open file ")+filename);
    else
      {
        char line[8096];
        while (file.getline(line, 8095))
          {
            char *l = line; while (*l && (*l != '=')) ++l;
            *(l-1) = 0;
            l++; while (*l == ' ') l++;
            insert(line, l);
          }
        file.close();
      }
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::exportFile(std::string filename) const
  {
    if (!db) return;
    ofstream file;
    file.open(filename.c_str(), ios::out);
    if (!file.good()) syslog(std::string("FileDatabase::exportFile() couldn't open file ")+filename);
    else
      {
        std::string key, value;
        int r = db->seq(db, &dbtKey, &dbtData, R_FIRST);
        while (!r)
          {
            key.assign((char*)dbtKey.data, dbtKey.size);
            value.assign((char*)dbtData.data, dbtData.size);
            file << key << " = " << value << endl;
            r = db->seq(db, &dbtKey, &dbtData, R_NEXT);
          }
        if (r < 0) syslog("FileDatabase::exportFile() seq() failed");
        file.flush();
        file.close();
      }
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::insert(std::string key, std::string value)
  {
    if (!db) return;
    toLower(key);
    dbtKey.size = key.length();
    dbtKey.data = (void*)key.c_str();
    dbtData.size = value.length();
    dbtData.data = (void*)value.c_str();
    int r = db->put(db, &dbtKey, &dbtData, 0);
    if (r) syslog("FileDatabase::insert() put() failed");
    return;
  }

// ----------------------------------------------------------------------------

void FileDatabase::remove(std::string key)
  {
    if (!db) return;
    toLower(key);
    dbtKey.size = key.length();
    dbtKey.data = (void*)key.c_str();
    int r = db->del(db, &dbtKey, 0);
    if (r < 0) syslog("FileDatabase::remove() del() failed");
    return;
  }

// ----------------------------------------------------------------------------

std::string FileDatabase::find(std::string key) const
  {
    if (!db) return "";
    toLower(key);
    std::string value;
    dbtKey.size = key.length();
    dbtKey.data = (void*)key.c_str();
    int r = db->get(db, &dbtKey, &dbtData, 0);
    if (!r) value.assign((char*)dbtData.data, dbtData.size);
    else if (r < 0) syslog("FileDatabase::find() get() failed");
    return value;
  }

// ----------------------------------------------------------------------------

void FileDatabase::apropos(std::vector<std::string> words, std::vector<std::string> &keys) const
  {
    if (!db) return;
    std::string key, txt;
    bool ok;
    int r = db->seq(db, &dbtKey, &dbtData, R_FIRST);
    while (!r)
      {
        txt.assign((char*)dbtKey.data, dbtKey.size);
        txt.append(" ");
        txt.append((char*)dbtData.data, dbtData.size);
        toLower(txt);
        ok = true;
        for (std::vector<std::string>::const_iterator j = words.begin(); ok && (j != words.end()); ++j)
          {
            if ((*j)[0] == '!') ok = (txt.find((*j).c_str()+1) == std::string::npos);
            else                ok = (txt.find((*j).c_str()  ) != std::string::npos);
          }
        if (ok)
          {
            key.assign((char*)dbtKey.data, dbtKey.size);
            keys.push_back(key);
          }
        r = db->seq(db, &dbtKey, &dbtData, R_NEXT);
      }
    if (r < 0) syslog("FileDatabase::apropos() seq() failed");
    return;
  }

#endif

// ----------------------------------------------------------------------------
