// IRC.cpp
//
// 20000818 / dhartmei - initial revision
// 20000830 / dhartmei - ownnick case-insensitive for PRIVMSG
// 20000905 / dhartmei - added clear/import/export database commands
// 20000909 / dhartmei - added new commands
// 20000920 / dhartmei - use USERHOST instead of WHOIS to get ident/host
// 20001001 / dhartmei - disabled FileDatabase for Linux compatibility
// 20001019 / dhartmei - reconnect after 15 mins of silence (lastRecv)
//

#include "IRC.hpp"
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <string>
#include <algorithm>

extern "C++" bool terminated;
extern "C++" char timestring[20];
extern "C++" void updateTime();
extern "C++" void syslog(std::string msg);

// ----------------------------------------------------------------------------

inline void toLower(std::string &s)
  {
    for (unsigned i = 0; i < s.length(); ++i)
      if ((s[i] >= 'A') && (s[i] <= 'Z')) s[i] = s[i]-'A'+'a';
    return;
  }

inline std::string replace(const std::string &s, char a, char b)
  {
    std::string t = s;
    for (unsigned i = 0; i < t.length(); ++i)
      if (t[i] == a) t[i] = b;
    return t;
  }

inline void replace(std::string &s, std::string k, std::string v)
  {
    unsigned p = s.find(k);
    while (p != std::string::npos)
      {
        s = s.substr(0, p) + v + s.substr(p+k.size(), s.size()-p-k.size());
        p = s.find(k);
      }
    return;
  }

inline void rtrim(std::string &s)
  {
    while (!s.empty() && *(s.end()-1) == ' ') s.erase(s.end()-1);
    return;
  }

void extract(std::string &s, std::string &t, std::string skip, std::string stop)
  {
    t = "";
    while (!s.empty() && (skip.find(*s.begin()) != std::string::npos)) s.erase(s.begin());
    while (!s.empty() && (stop.find(*s.begin()) == std::string::npos)) { t += s[0]; s.erase(s.begin()); }
    return;
  }

// ----------------------------------------------------------------------------

Rights::Rights(std::string filename) : filename(filename)
  {
    load();
  }

Rights::~Rights()
  {
    save();
  }

void Rights::load()
  {
    ifstream file;
    file.open(filename.c_str(), ios::in);
    if (file.good())
      {
        char a[1024];
        while (file.getline(a, 1023))
          {
            std::string line(a), type, mask;
            extract(line, type, " ", " "); toLower(type);
            extract(line, mask, " ", " ");
            if      (type == "oper"   ) opers.insert(mask);
            else if (type == "user"   ) users.insert(mask);
            else if (type == "partner")
              {
                std::string command; extract(line, command, " ", "");
                partners.insert(Partner(mask, command));
              }
          }
        file.close();
      }
    return;
  }

void Rights::save() const
  {
    ofstream file;
    file.open(filename.c_str(), ios::out);
    if (file.good())
      {
        for (PartnerSet::const_iterator i = partners.begin(); i != partners.end(); ++i)
          file << "partner " << (*i).mask << " " << (*i).command << endl;
        for (StringSet::const_iterator i = opers.begin(); i != opers.end(); ++i)
          file << "oper " << *i << endl;
        for (StringSet::const_iterator i = users.begin(); i != users.end(); ++i)
          file << "user " << *i << endl;
        file.flush();
        file.close();
      }
    return;
  }

bool Rights::isUser(const Person &person) const
  {
    return matches(person, users);
  }

bool Rights::isOper(const Person &person) const
  {
    return matches(person, opers);
  }

std::string Rights::isPartner(const Person &person) const
  {
    PartnerSet::const_iterator i = partners.begin();
    while ((i != partners.end()) && !matches(person, (*i).mask)) ++i;
    if (i == partners.end()) return ""; else return (*i).command;
  }

bool Rights::matches(const Person &person, const StringSet &set) const
  {
    bool match = false;
    for (StringSet::const_iterator i = set.begin(); !match && (i != set.end()); ++i)
      if (matches(person.nick, person.user, person.host, *i)) match = true;
    return match;
  }

void Rights::changePartner(std::string mask, std::string command)
  {
    Partner partner(mask, command);
    PartnerSet::iterator i = partners.find(partner);
    if (i != partners.end()) partners.erase(i);
    if (!command.empty())
      partners.insert(partner);
    return;
  }  

bool Rights::matches(const Person &person, std::string mask) const
  {
    return matches(person.nick, person.user, person.host, mask);
  }

bool Rights::matches(std::string nick, std::string user, std::string host, std::string mask) const
  {
    std::string nickmask, usermask, hostmask;
    extract(mask, nickmask, " ", "!");
    extract(mask, usermask, "! ", "@");
    extract(mask, hostmask, "@ ", " ");
    return matches(nick, nickmask) && matches(user, usermask) && matches(host, hostmask);
  }

bool Rights::matches(std::string s, std::string m) const
  {
    if (m.empty() || (m == "*")) return true;
    bool b = false, e = false;
    if (*m.begin() == '*') { b = true; m.erase(m.begin()); }
    if (*(m.end()-1) == '*') { e = true; m.erase(m.end()-1); }
    if (b && e) return s.find(m) != std::string::npos;
    else if (b) while (s.size() > m.size()) s.erase(s.begin());
    else if (e) while (s.size() > m.size()) s.erase(s.end()-1);
    return s == m;
  }

bool Rights::checkmask(std::string mask) const
  {
    unsigned a = mask.find_first_of('!'), b = mask.find_first_of('@');
    if ((a == std::string::npos) || (a != mask.find_last_of('!'))) return false;
    if ((b == std::string::npos) || (b != mask.find_last_of('@'))) return false;
    if (a > b) return false;
    if ((a == 0) || ((b-a) < 2) || (b == (mask.size()-1))) return false;
    return true;
  }

bool Rights::addUser(std::string mask)
  {
    if (users.find(mask) != users.end()) return false;
    else { users.insert(mask); return true; }
  }

bool Rights::addOper(std::string mask)
  {
    if (opers.find(mask) != opers.end()) return false;
    else { opers.insert(mask); return true; }
  }

bool Rights::removeUser(std::string mask)
  {
    if (users.find(mask) == users.end()) return false;
    else { users.erase(mask); return true; }
  }

bool Rights::removeOper(std::string mask)
  {
    if (opers.find(mask) == opers.end()) return false;
    else { opers.erase(mask); return true; }
  }

void Rights::listUsers(std::string &s) const
  {
    for (StringSet::const_iterator i = users.begin(); i != users.end(); ++i)
      { if (i != users.begin()) s += ", "; s += *i; }
    return;
  }

void Rights::listOpers(std::string &s) const
  {
    for (StringSet::const_iterator i = opers.begin(); i != opers.end(); ++i)
      { if (i != opers.begin()) s += ", "; s += *i; }
    return;
  }

void Rights::listPartners(std::string &s) const
  {
    for (PartnerSet::const_iterator i = partners.begin(); i != partners.end(); ++i)
      { if (i != partners.begin()) s += ", "; s += (*i).mask; s += " "; s += (*i).command; }
    return;
  }

// ----------------------------------------------------------------------------

Person::Person(std::string nick, std::string user, std::string host, bool op, bool voice) : nick(nick), user(user), host(host), op(op), voice(voice)
  {
  }

Person::~Person()
  {
  }

// ----------------------------------------------------------------------------

Channel::Channel(IRCSocket *ircsocket, std::string name) : ircsocket(ircsocket), name(name), op(false), rights(replace(name, '/', '_')+".cfg"), enableDictionary(ircsocket->ownnick == "Ambrose"), timeJoined(time(0))
  {
    file = new ofstream();
    file->open(replace(name, '/', '_').c_str(), ios::out|ios::app);
    log("################ channel log file openend #################");
#if (__WIN32__ || __Linux__)
    database = new MemoryDatabase(replace(name, '/', '_')+".dat");
#else
    database = new FileDatabase(replace(name, '/', '_')+".bin");
#endif
    database->load();
  }

Channel::~Channel()
  {
    while (!persons.empty())
      {
        Persons::iterator p = persons.begin();
        delete *p; persons.erase(p);
      }
    database->save();
    delete database;
    log("################# channel log file closed #################\n");
    file->flush(); file->close(); delete file; file = 0;
  }

void Channel::log(std::string msg)
  {
    if (file && file->good())
      {
        updateTime();
        *file << timestring << ' ' << msg << '\n';
        file->flush();
      }
    return;
  }

Person *Channel::findPerson(std::string nick)
  {
    Persons::iterator p = persons.begin();
    while ((p != persons.end()) && ((*p)->nick != nick)) p++;
    return (p == persons.end() ? 0 : *p);
  }

void Channel::join(Person &person, bool older, std::string ownnick)
  {
    if (!older)
      {
        char m[1024];
        sprintf(m, "--> %s (%s@%s) has joined %s", person.nick.c_str(), person.user.c_str(), person.host.c_str(), name.c_str());
        Channel::log(m);
      }
    persons.insert(persons.end(), &person);
    return;
  }

void Channel::part(Person &person, std::string user, std::string host)
  {
    char m[1024];
    sprintf(m, "<-- %s (%s@%s) has left %s", person.nick.c_str(), user.c_str(), host.c_str(), name.c_str());
    log(m);
    persons.erase(std::find(persons.begin(), persons.end(), &person));
    return;
  }

void Channel::quit(Person &person, std::string user, std::string host, std::string reason)
  {
    char m[1024];
    sprintf(m, "*** %s (%s@%s) Quit (%s)", person.nick.c_str(), user.c_str(), host.c_str(), reason.c_str());
    log(m);
    persons.erase(std::find(persons.begin(), persons.end(), &person));
    return;
  }

void Channel::kick(Person &person, std::string member, std::string reason)
  {
    char m[1024];
    sprintf(m, "*** %s was kicked by %s (%s)", person.nick.c_str(), member.c_str(), reason.c_str());
    log(m);
    persons.erase(std::find(persons.begin(), persons.end(), &person));
    return;
  }

void Channel::message(Person &person, std::string msg)
  {
    char m[1024];
    if (msg[0] == '\x01')
      {
        std::string command; extract(msg, command, "\x01 ", " "); toLower(command);
        std::string action; extract(msg, action, " ", "\x01");
        if (command == "action")
          sprintf(m, "* %s %s", person.nick.c_str(), action.c_str());
        else
          sprintf(m, "* %s %s %s", person.nick.c_str(), command.c_str(), action.c_str());
      }
    else
      sprintf(m, "<%s> %s", person.nick.c_str(), msg.c_str());
    log(m);
    if (enableDictionary) dictionary(person, msg);
    return;
  }

void Channel::dictionary(Person &person, std::string msg)
  {
    bool user  = person.op || rights.isOper(person) || rights.isUser(person);
    if (user)
      {
        std::string command; extract(msg, command, " ", " "); toLower(command);
        if      (command == "info")
          {
            std::string key; extract(msg, key, " ", "="); rtrim(key); toLower(key);
            if (!key.empty())
              {
                std::string value = database->find(key);
                if (value.empty()) send(std::string("\"")+key+"\" not found");
                else               send(key+" = "+value);
              }
          }
        else if (command == "rminfo")
          {
            std::string key; extract(msg, key, " ", "="); rtrim(key); toLower(key);
            if (!key.empty())
              {
                database->remove(key);
                send(std::string("\"")+key+"\" removed");
              }
          }
        else if (command == "mkinfo")
          {
            std::string key; extract(msg, key, " ", "="); rtrim(key); toLower(key);
            if (!key.empty())
              {
                std::string value; extract(msg, value, " =", "");
                if (!value.empty())
                  {
                    if (!database->find(key).empty())
                      send(std::string("\"")+key+"\" already defined");
                    else
                      {
                        database->insert(key, value);
                        send(std::string("\"")+key+"\" added");
                      }
                  }
              }
          }
        else if (command == "findinfo")
          {
            std::vector<std::string> words, keys;
            std::string key;
            extract(msg, key, " ", " "); toLower(key);
            while (!key.empty())
              {
                words.push_back(key);
                extract(msg, key, " ", " "); toLower(key);
              }
            database->apropos(words, keys);
            std::string result;
            if (keys.size() > 0)
              {
                char nr[10]; sprintf(nr, "%u", (unsigned)keys.size());
                result += nr;
                result += " matches: ";
                for (unsigned i = 0; (i < keys.size()) && (i < 10); ++i) { if (i > 0) { result += ", "; } result += keys[i]; }
                if (keys.size() > 10) result += " [...]";
              }
            else result = "no match";
            send(result);
          }
      }
    return;
  }

void Channel::nick(Person &person, std::string newnick)
  {
    char m[1024];
    sprintf(m, "--- %s is now known as %s", person.nick.c_str(), newnick.c_str());
    log(m);
    person.nick = newnick;
    return;
  }

void Channel::setTopic(std::string nick, std::string t)
  {
    char m[1024];
    sprintf(m, "--- %s has changed topic to: %s", nick.c_str(), t.c_str());
    log(m);
    topic = t;
    return;
  }

void Channel::mode(std::string nick, std::string modes, std::string arguments)
  {
    char m[1024]; sprintf(m, "--- %s changes mode %s %s", nick.c_str(), modes.c_str(), arguments.c_str()); log(m);
    bool set = false;
    while (!modes.empty())
      {
        char mode = modes[0]; modes.erase(modes.begin());
        if      (mode == '+') set = true;
        else if (mode == '-') set = false;
        else
          {
            std::string argument;
            if (( set && (std::string("bklov").find(mode) != std::string::npos)) ||
                (!set && (std::string("bkov" ).find(mode) != std::string::npos))   )
              extract(arguments, argument, " ", " ");
            if      (mode == 'o') // operator status
              {
                Person *p = findPerson(argument);
                if (!p) syslog(std::string("*** ERROR: MODE o for un-joined nick ")+argument);
                else
                  {
                    p->op = set;
                    if ((argument == ircsocket->ownnick) && (op != set))
                      {
                        op = set;
                        if (op) syslog(std::string("--- gained operator status in ")+name+" through "+nick);
                        else    syslog(std::string("--- lost operator status in ")+name+" through "+nick);
                      }
                  }
              }
            else if (mode == 'k') // key
              {
              }
            else if (mode == 'l') // limit
              {
              }
            else if (mode == 'b') // ban
              {
              }
            else if (mode == 'v') // voice
              {
                Person *p = findPerson(argument);
                if (!p) syslog(std::string("*** ERROR: MODE v for un-joined nick ")+argument);
                else p->voice = set;
              }
            else if (mode == 'i') // invite only
              {
              }
            else if (mode == 'm') // moderated
              {
              }
            else if (mode == 'n') // outside messages
              {
              }
            else if (mode == 'p') // private
              {
              }
            else if (mode == 's') // secret
              {
              }
            else if (mode == 't') // topic protected
              {
              }
            else syslog(std::string("*** ERROR: unsupported mode change in ")+name+" "+modes);
          }
      }
    return;
  }

void Channel::send(std::string msg)
  {
    ircsocket->sendLine(std::string("PRIVMSG ")+name+" :"+msg);
    log(std::string("<")+ircsocket->ownnick+"> "+msg);
    return;
  }

// ----------------------------------------------------------------------------

IRCSocket::IRCSocket(const char *address, unsigned port, const char *vhost, const char *nick) : LineSocket(address, port, vhost), ownnick(nick), reconnect(false)
  {
    loadRejoins();
  }

IRCSocket::~IRCSocket()
  {
    clearChannels();
    saveRejoins();
  }

void IRCSocket::clearChannels()
  {
    while (!channels.empty())
      {
        Channels::iterator c = channels.begin();
        delete *c; channels.erase(c);
      }
    return;
  }

bool IRCSocket::connect()
  {
    bool success = LineSocket::connect();
    if (success)
      {
        lastRecv = time(0);
        sendLine("PASS secret");
        sendLine(std::string("NICK ")+ownnick);
        sendLine("USER dhartmei 8 * :Daniel Hartmeier <daniel@benzedrine.cx>");
      }
    return success;
  }

bool IRCSocket::select(unsigned timeout)
  {
    unsigned long t = time(0);
    if (t > lastRecv+900) { reconnect = true; syslog("--- reconnecting (no data received for 15 minutes)"); }
    if (reconnect) { reconnect = false; return false; }

    // rejoin channels
    for (std::vector<Rejoin>::iterator i = rejoins.begin(), j = rejoins.end(); i != j; ++i)
      {
        Channel *c = findChannel((*i).channel);
        if (!c)
          {
            if ((*i).nextJoinRetry <= t)
              {
                syslog(std::string("--- trying to join ")+(*i).channel);
                sendLine(std::string("JOIN ")+(*i).channel);
                (*i).nextJoinRetry = t + 60;
                (*i).nextOpRequest = t + 5;
              }
          }
        else if (!c->op)
          {
            // request channel operator status
            if ((*i).nextOpRequest <= t)
              {
                std::vector<Partner> partners;
                for (Persons::iterator p = c->persons.begin(), q = c->persons.end(); p != q; ++p)
                  if ((*p)->op)
                    {
                      std::string command = c->rights.isPartner(**p);
                      if (!command.empty())
                        partners.push_back(Partner((*p)->nick, command));
                    }
                if (partners.size() > 0)
                  {
                    unsigned r = rand() % partners.size();
                    syslog(std::string("--- requesting operator status in ")+c->name+" from "+partners[r].mask);
                    replace(partners[r].command, "%c", c->name);
                    sendLine(std::string("PRIVMSG ")+partners[r].mask+" :"+partners[r].command);
                    (*i).nextOpRequest = t + 60;
                  }
                else (*i).nextOpRequest = t + 10;
              }
          }
      }

    return LineSocket::select(timeout);
  }

void IRCSocket::disconnect()
  {
    sendLine("QUIT");
    clearChannels();
    LineSocket::disconnect();
    return;
  }

Channel *IRCSocket::findChannel(std::string name)
  {
    Channels::iterator c = channels.begin();
    while ((c != channels.end()) && ((*c)->name != name)) c++;
    return (c == channels.end() ? 0 : *c);
  }

void IRCSocket::recvLine(std::string line)
  {
    std::string nick, user, host, command;
    if (line[0] == ':') extract(line, nick, ":", "! ");
    if (line[0] == '!') extract(line, user, "!", "@ ");
    if (line[0] == '@') extract(line, host, "@", " " );
    extract(line, command, " ", " ");

    if      (command == "PING")
      {
        // :irc.gigabell.de PING :irc.gigabell.de
        sendLine(std::string("PONG ")+ownnick);
      }
    else if (command == "JOIN")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch JOIN :#cryptonomicon
        std::string channel; extract(line, channel, " :", " "); toLower(channel);
        Channel *c = findChannel(channel);
        if (!c && (nick == ownnick)) { syslog(std::string("--- joined channel ")+channel); c = new Channel(this, channel); channels.insert(channels.end(), c); }
        if (!c) syslog(std::string("*** ERROR: JOIN for unknown channel ")+channel);
        else
          {
            Person *p = c->findPerson(nick);
            if (p) syslog(std::string("*** ERROR: JOIN for already joined nick ")+nick+" in "+channel);
            else
              {
                p = new Person(nick, user, host, false, false);
                c->join(*p, false, ownnick);
              }
          }
      }
    else if (command == "PART")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch PART #cryptonomicon
        std::string channel; extract(line, channel, " :", " "); toLower(channel);
        Channel *c = findChannel(channel);
        if (!c) syslog(std::string("*** ERROR: PART for unknown channel ")+channel);
        else
          {
            Person *p = c->findPerson(nick);
            if (!p) syslog(std::string("*** ERROR: PART for non-joined nick ")+nick+" in "+channel);
            else { c->part(*p, user, host); delete p; }
            if (nick == ownnick) { syslog(std::string("--- left channel ")+channel); channels.erase(std::find(channels.begin(), channels.end(), c)); delete c; }
          }
      }
    else if (command == "QUIT")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch QUIT :Gone to have lunch
        std::string reason; extract(line, reason, " :", "");
        for (Channels::iterator c = channels.begin(); c != channels.end(); ++c)
          {
            Person *p = (*c)->findPerson(nick);
            if (p) { (*c)->quit(*p, user, host, reason); delete p; }
          }
        if (nick == ownnick) reconnect = true;
      }
    else if (command == "KICK")
      {
        // dhartmei!dhartmei@cable-ggar48-183.intergga.ch KICK  #cryptonomicon Ambrose :nick flood
        std::string channel; extract(line, channel, " ", " "); toLower(channel);
        std::string member; extract(line, member, " ", " ");
        std::string reason; extract(line, reason, " :", "");
        Channel *c = findChannel(channel);
        if (!c) syslog(std::string("*** ERROR: KICK for unknown channel ")+channel);
        else
          {
            Person *p = c->findPerson(member);
            if (!p) syslog(std::string("*** ERROR: KICK for non-joined nick ")+member+" in "+channel);
            else { c->kick(*p, nick, reason); delete p; }
            if (member == ownnick)
              {
                syslog(std::string("--- kicked from channel ")+channel+" by "+nick+" ("+reason+")");
                channels.erase(std::find(channels.begin(), channels.end(), c));
                delete c;
                std::vector<Rejoin>::iterator i = std::find(rejoins.begin(), rejoins.end(), Rejoin(channel));
                if (i != rejoins.end()) { (*i).nextJoinRetry = time(0)+10; (*i).nextOpRequest = 0; }
              }
          }
      }
    else if (command == "PRIVMSG")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch PRIVMSG #cryptonomicon :channel message
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch PRIVMSG Ambrose :private message
        std::string target; extract(line, target, " ", " "); toLower(target);
        std::string message; extract(line, message, " :", "");
        if (!target.empty() && !message.empty())
          {
            if (target[0] == '#')
              {
                Channel *c = findChannel(target);
                if (!c) syslog(std::string("*** ERROR: PRIVMSG for unknown channel ")+target+" "+message);
                else
                  {
                    Person *p = c->findPerson(nick);
                    if (!p) syslog(std::string("*** ERROR: PRIVMSG from non-joined nick ")+nick+" in "+target+" "+message);
                    else { p->user = user; p->host = host; c->message(*p, message); }
                  }
              }
            else
              {
                std::string myself = ownnick; toLower(myself);
                if (target != myself) syslog(std::string("*** ERROR: PRIVMSG for non-ownnick! ")+target+" "+nick+"!"+user+"@"+host+" "+message);
                else
                  {
                    if (message[0] == '\x01')
                      ctcp(nick, user, host, message);
                    else
                      execute(nick, user, host, message);
                  }
              }
          }
      }
    else if (command == "NICK")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch NICK :dhartmei2
        std::string newnick; extract(line, newnick, " :", " ");
        for (Channels::iterator c = channels.begin(); c != channels.end(); ++c)
          {
            Person *p = (*c)->findPerson(nick);
            if (p) (*c)->nick(*p, newnick);
          }
        if (nick == ownnick) ownnick = newnick;
      }
    else if (command == "TOPIC")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch TOPIC #cryptonomicon :new topic
        std::string channel; extract(line, channel, " ", " "); toLower(channel);
        std::string topic; extract(line, topic, " :", "");
        Channel *c = findChannel(channel);
        if (!c) syslog(std::string("*** ERROR: TOPIC for unknown channel ")+channel+" "+topic);
        else c->setTopic(nick, topic);
      }
    else if (command == "MODE")
      {
        // :dhartmei!dhartmei@cable-ggar48-183.intergga.ch MODE #cryptonomicon +ooo hardtime mathride midearth
        std::string channel; extract(line, channel, " ", " "); toLower(channel);
        std::string modes; extract(line, modes, " ", " ");
        std::string arguments; extract(line, arguments, " ", "");
        Channel *c = findChannel(channel);
        if (!c) syslog(std::string("*** ERROR: MODE for unknown channel ")+channel+" "+modes+" "+arguments);
        else c->mode(nick, modes, arguments);
      }
    else if (command == "332")
      {
        // :irc.gigabell.de 332 dhartmei2 #cryptonomicon :new topic
        std::string receiver; extract(line, receiver, " ", " ");
        std::string channel; extract(line, channel, " ", " "); toLower(channel);
        std::string topic; extract(line, topic, " :", "");
        Channel *c = findChannel(channel);
        if (!c) syslog(std::string("*** ERROR: 332 for unknown channel ")+channel);
        else c->setTopic(nick, topic);
      }
    else if (command == "353") // reply to JOIN command
      {
        // :irc.gigabell.de 353 dhartmei = #cryptonomicon :dhartmei @mathride @hardtime @midearth
        // :irc.gigabell.de 353 dhartmei * #cryptonomicon :dhartmei @mathride @hardtime @midearth
        // :irc.gigabell.de 353 dhartmei @ #cryptonomicon :dhartmei @mathride @hardtime @midearth
        std::string self; extract(line, self, " ", " ");
        std::string channel; extract(line, channel, " =*@", " "); toLower(channel);
        std::string userlist;
        unsigned count = 0;
        while (!line.empty())
          {
            std::string member; extract(line, member, " :", " ");
            if (!member.empty())
              {
                bool op = false, voice = false;
                if (member[0] == '@') { op    = true; member.erase(member.begin()); }
                if (member[0] == '+') { voice = true; member.erase(member.begin()); }
                if (member != ownnick)
                  {
                    Channel *c = findChannel(channel);
                    if (!c) syslog(std::string("*** ERROR: 353 for unknown channel ")+channel);
                    else
                      {
                        Person *p = c->findPerson(member);
                        if (p) syslog(std::string("*** ERROR: 353 for already joined nick ")+member+" in "+channel);
                        else
                          {
                            p = new Person(member, "", "", op, voice);
                            c->join(*p, true, ownnick);
                          //syslog(std::string("--- requesting WHOIS of ")+member);
                          //sendLine(std::string("WHOIS ")+member);
                            if (count) userlist += ' '; userlist += member;
                            if (++count == 5)
                              {
                                syslog(std::string("--- requesting USERHOST ")+userlist);
                                sendLine(std::string("USERHOST ")+userlist);
                                userlist = ""; count = 0;
                              }
                          }
                      }
                  }
              }
          }
        if (count)
          {
            syslog(std::string("--- requesting USERHOST ")+userlist);
            sendLine(std::string("USERHOST ")+userlist);
          }
      }
    else if (command == "302") // reply to USERHOST command
      {
        // irc.gigabell.de 302 Ambrose :dhartmei=+~dhartmei@openbsd.reichardt.ch Dengo=+dhartmei@openbsd.nightlights.ch
        std::string self; extract(line, self, " ", " ");
        while (!line.empty())
          {
            std::string member; extract(line, member, " :", " ");
            if (!member.empty())
              {
                // dhartmei=+~dhartmei@openbsd.reichardt.ch
                std::string nick; extract(member, nick, "", "=");
                std::string user; extract(member, user, "=", "@");
                std::string host; extract(member, host, "@", "");
                if (nick.empty() || user.empty() || host.empty()) syslog("*** ERROR: 302 with invalid data");
                else
                  {
                    bool ircop = false, away = false;
                    if (*(nick.end()-1) == '*') { ircop = true; nick.erase(nick.end()-1); }
                    if ((user[0] == '+') || (user[0] == '-'))
                      { away = (user[0] == '-'); user.erase(user.begin()); }
                    syslog(std::string("--- received USERHOST reply ")+nick+"!"+user+"@"+host+(ircop ? " IRCOP" : "")+(away ? " AWAY" : ""));
                    for (Channels::iterator c = channels.begin(); c != channels.end(); ++c)
                      {
                        Person *p = (*c)->findPerson(nick);
                        if (p) { p->user = user; p->host = host; }
                      }
                  }
              }
          }
      }
  //else if (command == "311") { /* WHOIS -> user */ }
  //else if (command == "312") { /* WHOIS -> server */ }
  //else if (command == "317") { /* WHOIS -> idle time */ }
  //else if (command == "318") { /* WHOIS -> END */ }
  //else if (command == "319") { /* WHOIS -> channels */ }
    else if (command == "333") { /* channel creation time on JOIN */ }
    else if (command == "366") { /* end of NAMES list */ }
    else if (command == "372") { /* MOTD */ }
    else if (command == "375") { /* begin of MOTD command */ }
    else if (command == "376") { /* end of MOTD command */ }
    else syslog(std::string("*** ")+nick+"!"+user+"@"+host+" "+command+line);

    lastRecv = time(0);
    return;
  }

void IRCSocket::ctcp(std::string nick, std::string user, std::string host, std::string message)
  {
    std::string command; extract(message, command, "\x01 ", " \x01"); toLower(command);
    if      (command == "version")
      {
        syslog(std::string("--- answering CTCP VERSION from ")+nick+"!"+user+"@"+host);
        sendLine(std::string("NOTICE ")+nick+" :\x01VERSION \"InfoBot v1.23\" \"OpenBSD 2.7 (i386)\" \"Daniel Hartmeier <daniel@benzedrine.cx>\"\x01");
      }
    else if (command == "ping")
      {
        std::string argument; extract(message, argument, " ", "\x01"); toLower(argument);
        syslog(std::string("--- answering CTCP PING ")+argument+" from "+nick+"!"+user+"@"+host);
        if (argument.length() <= 16)
          sendLine(std::string("NOTICE ")+nick+" :\x01PING "+argument+"\x01");
      }
    else syslog(std::string("--- received unsupported CTCP ")+command+" from "+nick+"!"+user+"@"+host);
    return;
  }

void IRCSocket::execute(std::string nick, std::string user, std::string host, std::string message)
  {
    syslog(std::string("--- PRIVMSG from ")+nick+"!"+user+"@"+host+" "+message);
    bool owner = (nick == "dhartmei") && ((host == "openbsd.reichardt.ch") || (host == "openbsd.nightlights.ch") || (host == "cable-ggar48-183.intergga.ch"));
    std::string command; extract(message, command, " ", " "); toLower(command);
    if (command == "help")
      {
        sendLine(std::string("NOTICE ")+nick+" :channel <channel> status / shutup / answer / list / grant|deny user|oper <mask> / op / topic <text> / voice <nick> / devoice <nick> / kick <nick> <reason> / kickban <mask> <reason> / unban <mask>");
        sendLine(std::string("NOTICE ")+nick+" :for a full explanation of all commands, see http://www.benzedrine.cx/ambrose.html");
        syslog(std::string("--- ")+nick+" requests help");
      }
    else if (command == "channel")
      {
        std::string channel; extract(message, channel, " ", " "); toLower(channel);
        Channel *c = findChannel(channel);
        if (c)
          {
            Person *p = c->findPerson(nick);
            if (owner || (p && c->rights.isOper(*p)))
              {
                if (command == "channel")
                  {
                    std::string subcommand; extract(message, subcommand, " ", " "); toLower(subcommand);
                    if (subcommand == "status")
                      {
                        sendLine(std::string("NOTICE ")+nick+" :mode is "+(c->enableDictionary ? "answer" : "quiet"));
                        syslog(std::string("--- ")+channel+" "+nick+" requests status");
                      }
                    else if (subcommand == "shutup")
                      {
                        c->enableDictionary = false;
                        sendLine(std::string("NOTICE ")+nick+" :mode set to quiet");
                        syslog(std::string("--- ")+channel+" "+nick+" sets mode to quiet");
                      }
                    else if (subcommand == "answer")
                      {
                        c->enableDictionary = true;
                        sendLine(std::string("NOTICE ")+nick+" :mode set to answer");
                        syslog(std::string("--- ")+channel+" "+nick+" sets mode to answer");
                      }
                    else if (subcommand == "partner")
                      {
                        if (owner)
                          {
                            std::string mask; extract(message, mask, " ", " ");
                            std::string command; extract(message, command, " ", "");
                            if (mask.empty())
                              {
                                std::string m; c->rights.listPartners(m);
                                sendLine(std::string("NOTICE ")+nick+" :partners: "+m);
                                syslog(std::string("--- ")+channel+" "+nick+" lists partners");
                              }
                            else
                              {
                                syslog(std::string("--- ")+channel+" "+nick+" changes partner "+mask+" "+command);
                                c->rights.changePartner(mask, command);
                                c->rights.save();
                                sendLine(std::string("NOTICE ")+nick+" :changed partner "+mask+" "+command);
                              }
                          }
                      }
                    else if (subcommand == "grant")
                      {
                        std::string type; extract(message, type, " ", " "); toLower(type);
                        std::string mask; extract(message, mask, " ", " ");
                        if (type == "user")
                          {
                            if      (!c->rights.checkmask(mask)) sendLine(std::string("NOTICE ")+nick+" :invalid mask "+mask+" (try *nick*!*user*@*host*)");
                            else if (!c->rights.addUser(mask)  ) sendLine(std::string("NOTICE ")+nick+" :already user "+mask);
                            else
                              {
                                c->rights.save();
                                sendLine(std::string("NOTICE ")+nick+" :granted user access to "+mask);
                                syslog(std::string("--- ")+channel+" "+nick+" grants user access to "+mask);
                              }
                          }
                        else if (type == "oper")
                          {
                            if      (!c->rights.checkmask(mask)) sendLine(std::string("NOTICE ")+nick+" :invalid mask "+mask+" (try *nick*!*user*@*host*)");
                            else if (!c->rights.addOper(mask)  ) sendLine(std::string("NOTICE ")+nick+" :already oper "+mask);
                            else
                              {
                                c->rights.save();
                                sendLine(std::string("NOTICE ")+nick+" :granted oper access to "+mask);
                                syslog(std::string("--- ")+channel+" "+nick+" grants oper access to "+mask);
                              }
                          }
                      }
                    else if (subcommand == "deny")
                      {
                        std::string type; extract(message, type, " ", " "); toLower(type);
                        std::string mask; extract(message, mask, " ", " ");
                        if (type == "user")
                          {
                            if      (!c->rights.checkmask(mask) ) sendLine(std::string("NOTICE ")+nick+" :invalid mask "+mask+" (try *nick*!*user*@*host*)");
                            else if (!c->rights.removeUser(mask)) sendLine(std::string("NOTICE ")+nick+" :no user "+mask);
                            else
                              {
                                c->rights.save();
                                sendLine(std::string("NOTICE ")+nick+" :denied user access to "+mask);
                                syslog(std::string("--- ")+channel+" "+nick+" denies user access to "+mask);
                              }
                          }
                        else if (type == "oper")
                          {
                            if      (!c->rights.checkmask(mask) ) sendLine(std::string("NOTICE ")+nick+" :invalid mask "+mask+" (try *nick*!*user*@*host*)");
                            else if (!c->rights.removeOper(mask)) sendLine(std::string("NOTICE ")+nick+" :no oper "+mask);
                            else
                              {
                                c->rights.save();
                                sendLine(std::string("NOTICE ")+nick+" :denied oper access to "+mask);
                                syslog(std::string("--- ")+channel+" "+nick+" denies oper access to "+mask);
                              }
                          }
                      }
                    else if (subcommand == "list")
                      {
                        std::string m;
                        m = "NOTICE "+nick+" :opers: ";
                        c->rights.listOpers(m);
                        sendLine(m);
                        m = "NOTICE "+nick+" :users: ";
                        c->rights.listUsers(m);
                        sendLine(m);
                        syslog(std::string("--- ")+channel+" "+nick+" lists access");
                      }
                    else if (subcommand == "op")
                      {
                        if (p && !p->op)
                          {
                            sendLine(std::string("MODE ")+channel+" +o "+nick);
                            syslog(std::string("--- giving channel operator status to ")+nick+"!"+user+"@"+host+" in "+channel);
                          }
                      }
                    else if (subcommand == "topic")
                      {
                        std::string topic; extract(message, topic, " ", "");
                        sendLine(std::string("TOPIC ")+channel+" :"+topic);
                      }
                    else if (subcommand == "voice")
                      {
                        std::string target; extract(message, target, " ", " ");
                        while (!target.empty())
                          {
                            Person *p = c->findPerson(target);
                            if (p && !p->voice)
                              sendLine(std::string("MODE ")+channel+" +v "+target);
                            extract(message, target, " ", " ");
                          }
                      }
                    else if (subcommand == "devoice")
                      {
                        std::string target; extract(message, target, " ", " ");
                        while (!target.empty())
                          {
                            Person *p = c->findPerson(target);
                            if (p && p->voice)
                              sendLine(std::string("MODE ")+channel+" -v "+target);
                            extract(message, target, " ", " ");
                          }
                      }
                    else if (subcommand == "kick")
                      {
                        std::string target; extract(message, target, " ", " ");
                        std::string reason; extract(message, reason, " ", "");
                        if (reason.empty()) reason = "requested";
                        Person *p = c->findPerson(target);
                        if (p && !c->rights.isOper(*p))
                          sendLine(std::string("KICK ")+channel+" "+target+" :"+reason);
                      }
                    else if (subcommand == "kickban")
                      {
                        std::string mask; extract(message, mask, " ", " ");
                        if (mask.find('!') == std::string::npos) mask += "!*@*";
                        std::string reason; extract(message, reason, " ", "");
                        if (reason.empty()) reason = "requested";
                        sendLine(std::string("MODE ")+channel+" +b "+mask);
                        for (Persons::iterator i = c->persons.begin(); i != c->persons.end(); ++i)
                          if (c->rights.matches(**i, mask) && !c->rights.isOper(**i))
                            sendLine(std::string("KICK ")+channel+" "+(*i)->nick+" :"+reason);
                      }
                    else if (subcommand == "unban")
                      {
                        std::string mask; extract(message, mask, " ", " ");
                        if (!mask.empty())
                          sendLine(std::string("MODE ")+channel+" -b "+mask);
                      }
                    else if (subcommand == "save")
                      {
                        if (owner)
                          {
                            c->database->save();
                            sendLine(std::string("NOTICE ")+nick+" :saved database "+c->name);
                          }
                      }
                    else if (subcommand == "clear")
                      {
                        if (owner)
                          {
                            c->database->clear();
                            sendLine(std::string("NOTICE ")+nick+" :cleared database "+c->name);
                          }
                      }
                    else if (subcommand == "import")
                      {
                        std::string file; extract(message, file, " ", " ");
                        if (owner && !file.empty())
                          {
                            c->database->importFile(file);
                            sendLine(std::string("NOTICE ")+nick+" :imported file "+file+" into database "+c->name);
                          }
                      }
                    else if (subcommand == "export")
                      {
                        std::string file; extract(message, file, " ", " ");
                        if (owner && !file.empty())
                          {
                            c->database->exportFile(file);
                            sendLine(std::string("NOTICE ")+nick+" :exported file "+file+" from database "+c->name);
                          }
                      }
                  }
              }
          }
      }
    else if (command == "debug")
      {
        if (owner)
          {
            std::string r;
            for (std::vector<Rejoin>::const_iterator i = rejoins.begin(); i != rejoins.end(); ++i)
              { if (i != rejoins.begin()) r += ", "; r += (*i).channel; }
            sendLine(std::string("NOTICE ")+nick+" :rejoin list: "+r);
            if (!channels.size()) sendLine(std::string("NOTICE ")+nick+" :no channels joined");
            for (Channels::iterator c = channels.begin(); c != channels.end(); ++c)
              {
                unsigned ops = 0;
                for (Persons::iterator p = (*c)->persons.begin(); p != (*c)->persons.end(); ++p)
                  if ((*p)->op) ops++;
                char m[1024];
                sprintf(m, "channel %s (%u ops, %u others), mode %s", (*c)->name.c_str(), ops, (unsigned)(*c)->persons.size()-ops, ((*c)->enableDictionary ? "answer" : "quiet"));
                sendLine(std::string("NOTICE ")+nick+" :"+m);
              }
          }
      }
    else if (command == "rejoin")
      {
        if (owner)
          {
            std::string subcommand; extract(message, subcommand, " ", " "); toLower(subcommand);
            std::string channel; extract(message, channel, " ", " "); toLower(channel);
            if (subcommand == "add")
              {
                std::vector<Rejoin>::iterator i = std::find(rejoins.begin(), rejoins.end(), Rejoin(channel));
                if (i != rejoins.end())
                  sendLine(std::string("NOTICE ")+nick+" :channel "+channel+" already on rejoin list");
                else
                  {
                    rejoins.push_back(Rejoin(channel));
                    saveRejoins();
                    sendLine(std::string("NOTICE ")+nick+" :channel "+channel+" added to rejoin list");
                  }
              }
            else if (subcommand == "remove")
              {
                std::vector<Rejoin>::iterator i = std::find(rejoins.begin(), rejoins.end(), Rejoin(channel));
                if (i == rejoins.end())
                  sendLine(std::string("NOTICE ")+nick+" :channel "+channel+" not on rejoin list");
                else
                  {
                    rejoins.erase(i);
                    saveRejoins();
                    sendLine(std::string("NOTICE ")+nick+" :channel "+channel+" removed rejoin list");
                  }
              }
          }
      }
    else if (command == "relay")
      {
        if (owner)
          {
            std::string relay; extract(message, relay, " ", "");
            syslog(std::string("--- relaying command \"")+relay+"\"");
            sendLine(std::string("NOTICE ")+nick+" :relayed "+relay);
            sendLine(relay);
          }
      }
    else if (command == "terminate")
      {
        if (owner)
          {
            syslog("--- termination set");
            terminated = true; reconnect = true;
            sendLine(std::string("NOTICE ")+nick+" :termination set");
          }
      }
    return;
  }

void IRCSocket::loadRejoins()
  {
    rejoins.erase(rejoins.begin(), rejoins.end());
    ifstream file;
    file.open("InfoBot.channels", ios::in);
    if (file.good())
      {
        char a[1024];
        while (file.getline(a, 1023))
          {
            std::string line(a); toLower(line); rtrim(line);
            if ((line.length() > 1) && (line[0] == '#'))
              rejoins.push_back(Rejoin(line));
          }
        file.close();
      }
    return;
  }

void IRCSocket::saveRejoins() const
  {
    ofstream file;
    file.open("InfoBot.channels", ios::out);
    if (file.good())
      {
        for (std::vector<Rejoin>::const_iterator i = rejoins.begin(); i != rejoins.end(); ++i)
          file << (*i).channel << endl;
        file.flush(); file.close();
      }
    return;
  }

// ----------------------------------------------------------------------------

