// IRC.hpp
//
// 20000818 / dhartmei - initial revision
//

#ifndef IRC_HPP
#define IRC_HPP

#include "LineSock.hpp"
#include "Database.hpp"
#include <string>
#include <vector>
#include <set>
#include <fstream>

// ----------------------------------------------------------------------------

class Person;
class Rights;
class Channel;
class IRCSocket;

// ----------------------------------------------------------------------------

class Person
  {
    public:
      Person(std::string nick, std::string user, std::string host, bool op, bool voice);
      virtual ~Person();
      virtual bool operator!=(const Person &other) const { return nick != other.nick; }
      std::string nick, user, host;
      bool op, voice;
  };

typedef std::vector<Person*> Persons;

// ----------------------------------------------------------------------------

class Partner
  {
    public:
      Partner() {}
      Partner(std::string mask, std::string command) : mask(mask), command(command) {}
      virtual ~Partner() {}
      virtual bool operator<(const Partner &other) const { return mask < other.mask; }
      std::string mask, command;
  };

typedef std::set< Partner, std::less<Partner> > PartnerSet;

// ----------------------------------------------------------------------------

class Rights
  {
    public:
      Rights(std::string filename);
      virtual ~Rights();
      virtual void load();
      virtual void save() const;
      virtual bool isUser(const Person &person) const;
      virtual bool isOper(const Person &person) const;
      virtual std::string isPartner(const Person &person) const;
      virtual bool checkmask(std::string mask) const;
      virtual bool addUser(std::string mask);
      virtual bool addOper(std::string mask);
      virtual bool removeUser(std::string mask);
      virtual bool removeOper(std::string mask);
      virtual void changePartner(std::string mask, std::string command);
      virtual void listUsers(std::string &s) const;
      virtual void listOpers(std::string &s) const;
      virtual void listPartners(std::string &s) const;
      virtual bool matches(const Person &person, std::string mask) const;
    protected:
      typedef std::set< std::string, std::less<std::string> > StringSet;
      virtual bool matches(const Person &person, const StringSet &set) const;
      virtual bool matches(std::string nick, std::string user, std::string host, std::string mask) const;
      virtual bool matches(std::string s, std::string m) const;
      std::string filename;
      StringSet users, opers;
      PartnerSet partners;
  };

// ----------------------------------------------------------------------------

class Channel
  {
    public:
      Channel(IRCSocket *ircsocket, std::string name);
      virtual ~Channel();
      virtual bool operator!=(const Channel &other) const { return name != other.name; }
      virtual void log(std::string msg);
      virtual void join(Person &person, bool older, std::string ownnick);
      virtual void part(Person &person, std::string user, std::string host);
      virtual void quit(Person &person, std::string user, std::string host, std::string reason);
      virtual void kick(Person &person, std::string member, std::string reason);
      virtual void message(Person &person, std::string msg);
      virtual void dictionary(Person &person, std::string msg);
      virtual void nick(Person &person, std::string newnick);
      virtual void setTopic(std::string nick, std::string t);
      virtual void mode(std::string nick, std::string modes, std::string arguments);
      virtual void send(std::string msg);
      Person *findPerson(std::string nick);
      IRCSocket *ircsocket;
      std::string name, topic;
      Persons persons;
      ofstream *file;
      bool op;
      Database *database;
      Rights rights;
      bool enableDictionary;
      unsigned long timeJoined;
  };

typedef std::vector<Channel*> Channels;

// ----------------------------------------------------------------------------

class IRCSocket : public LineSocket
  {
    public:
      IRCSocket(const char *address, unsigned port, const char *vhost, const char *nick);
      virtual ~IRCSocket();
      virtual void clearChannels();
      virtual bool select(unsigned timeout = 0);
      virtual bool connect();
      virtual void disconnect();
      virtual void recvLine(std::string line);
      virtual void ctcp(std::string nick, std::string user, std::string host, std::string message);
      virtual void execute(std::string nick, std::string user, std::string host, std::string message);
      Channel *findChannel(std::string name);
      std::string ownnick;
      Channels channels;
      bool reconnect;
      class Rejoin
        {
          public:
            Rejoin() {}
            Rejoin(std::string channel) : channel(channel), nextJoinRetry(0), nextOpRequest(0) {}
            bool operator==(const Rejoin &other) const { return channel == other.channel; }
            bool operator!=(const Rejoin &other) const { return channel != other.channel; }
            std::string channel;
            unsigned long nextJoinRetry, nextOpRequest;
        };
      std::vector<Rejoin> rejoins;
      void loadRejoins();
      void saveRejoins() const;
      unsigned long lastRecv;
  };

// ----------------------------------------------------------------------------

#endif

