// LineSock.cpp
//
// 20000818 / dhartmei - initial revision
// 20001001 / dhartmei - Linux compatibility issues
// 20001017 / dhartmei - added bind() for vhost support
//

#include "LineSock.hpp"

#ifdef __WIN32__
  #include <winsock2.h>
  typedef unsigned long in_addr_t;
  typedef unsigned long ssize_t;
  WSADATA wsadata;
#else
  #include <sys/types.h>
  #include <sys/time.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
  #include <unistd.h>
  #include <netdb.h>
  #ifdef __Linux__
    typedef unsigned long in_addr_t;
  #endif
#endif
#include <errno.h>
#include <string.h>
#include <string>

extern "C++" void syslog(std::string msg);

// ----------------------------------------------------------------------------

bool InitializeSockets()
  {
    bool success = true;
#ifdef __WIN32__
    if (WSAStartup(MAKEWORD(2, 0), &wsadata))
      {
        syslog("InitializeSockets() WSAStartup() failed");
        success = false;
      }
#endif
    return success;
  }

// ----------------------------------------------------------------------------

void CleanupSockets()
  {
#ifdef __WIN32__
    if (WSACleanup())
      syslog("CleanupSockets() WSACleanup() failed");
#endif
    return;
  }

// ----------------------------------------------------------------------------

StreamSocket::StreamSocket(const char *address, unsigned port, const char *vhost)
  {
    this->address = new char[strlen(address)+1];
    strcpy(this->address, address);
    this->port = port;
    if (vhost && *vhost)
      {
        this->vhost = new char[strlen(vhost)+1];
        strcpy(this->vhost, vhost);
      }
    else
      {
        this->vhost = new char[1024];
        if (gethostname(this->vhost, 1024)) syslog("StreamSocket::StreamSocket() gethostname() failed");
      }
    s = 0;
  }

StreamSocket::~StreamSocket()
  {
    delete [] vhost;
    delete [] address;
  }

bool fillsin(sockaddr_in &sin, unsigned port, const char *address)
  {
    bool success = false;
    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port   = htons(port);
    in_addr_t a = inet_addr(address);
    if (a == INADDR_NONE)
      {
        syslog(std::string("resolving host name ")+address);
        struct hostent *h = gethostbyname(address);
        if (!h) syslog(std::string("invalid host name ")+address);
        else memcpy(&a, h->h_addr, sizeof(a));
      }
    if (a != INADDR_NONE)
      {
        memcpy(&sin.sin_addr, &a, 4);
        success = true;
      }
    return success;
  }

bool StreamSocket::connect()
  {
    bool success = false;
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s == -1) syslog("StreamSocket::connect() socket() failed");
    else
      {
        struct sockaddr_in sin;
        if (fillsin(sin, 0, vhost))
          {
            syslog(std::string("binding to ")+inet_ntoa(sin.sin_addr));
            if (bind(s, (sockaddr*)&sin, sizeof(sockaddr_in)))
              syslog("StreamSocket::connect() bind() failed");
            else
              {
                if (fillsin(sin, port, address))
                  { 
                    syslog(std::string("connecting to ")+inet_ntoa(sin.sin_addr));
                    if (::connect(s, (sockaddr*)&sin, sizeof(sockaddr_in)))
                      syslog("StreamSocket::connect() connect() failed");
                    else
                      success = true;
                  }
              }
          }
      }
    return success;
  }

void StreamSocket::disconnect()
  {
#ifdef __WIN32__
    int r = shutdown(s, SD_BOTH);
#else
    int r = shutdown(s, SHUT_RDWR);
#endif
    if (r) syslog("StreamSocket::disconnect() shutdown() failed");
    else s = 0;
    return;
  }

bool StreamSocket::select(unsigned timeout)
  {
    bool success = true;
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(s, &readfds);
    timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = timeout;
    int r = ::select(s+1, &readfds, 0, 0, &tv);
    if (r > 0)
      {
        char buf[4096];
        ssize_t n = ::recv(s, buf, 4095, 0); buf[n] = 0;
        if (n <= 0) { syslog("StreamSocket::select() recv() failed"); success = false; }
        else recv(buf);
      }
    return success;
  }

void StreamSocket::send(const char *data)
  {
    ssize_t l = strlen(data);
    ssize_t n = ::send(s, data, l, 0);
    if (n != l) syslog("StreamSocket::send() send() failed");
    return;
  }

// ----------------------------------------------------------------------------

LineSocket::LineSocket(const char *address, unsigned port, const char *vhost) : StreamSocket(address, port, vhost)
  {
  }

void LineSocket::recv(const char *data)
  {
    while (*data)
      {
             if (*data == '\n') { recvLine(line); line = ""; }
        else if (*data != '\r')   line += *data;
        data++;
      }
    return;
  }

void LineSocket::sendLine(std::string line)
  {
    line += '\n';
    send(line.c_str());
    return;
  }

// ----------------------------------------------------------------------------

